import io
import os
from types import TracebackType
from typing import Optional, Type

import boto3
from botocore.exceptions import ClientError

from .fs import FS, FileStat


class S3ObjectStat(FileStat):
    def __init__(self, key, head):
        self.filename = key
        self.last_modified = head['LastModified'].timestamp()
        self.size = head['ContentLength']
        self.metadata = head['Metadata']
        self._head = head

    def isdir(self):
        return False


class _ObjectReader:
    def __init__(self, client, bucket, key, mode, kwargs):
        self.client = client
        self.res = self.client.get_object(Bucket=bucket,
                                          Key=key)
        self._mode = mode
        self.body = self.res['Body']
        self.pos = 0
        self.content_length = self.res['ContentLength']

    def read(self, size=-1):
        if size <= 0:
            size = None

        if self.content_length <= self.pos:
            return

        if 'b' in self._mode:
            data = self.body.read(size)
        else:
            data = self.body.read(size).decode('utf-8')

        self.pos += len(data)
        return data

    def close(self):
        self.body = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]],
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> bool:
        self.close()

    def flush(self):
        pass

    @property
    def closed(self):
        return self.body is None

    def isatty(self):
        return False

    def readable(self):
        return True

    def seekable(self):
        return False

    def writable(self):
        return False


class _ObjectWriter:
    def __init__(self, client, bucket, key, mode, kwargs):
        self.client = client
        self.bucket = bucket
        self.key = key
        if 'b' in mode:
            self.buf = io.BytesIO()
        else:
            self.buf = io.StringIO()

    def write(self, buf):
        return self.buf.write(buf)

    def close(self):
        # TODO: MPU
        # See:  https://boto3.amazonaws.com/v1/documentation/
        # api/latest/reference/services/s3.html#S3.Client.put_object
        self.client.put_object(Body=self.buf.getvalue(),
                               Bucket=self.bucket,
                               Key=self.key)
        self.buf = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]],
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> bool:
        self.close()

    def flush(self):
        '''Does nothing
        '''
        pass

    @property
    def closed(self):
        return self.buf is None

    def isatty(self):
        return False

    def readable(self):
        return False

    def seekable(self):
        return False

    def writable(self):
        return True


class S3(FS):
    '''S3 FileSystem wrapper

    Takes three arguments as well as enviroment variables for
    constructor. The priority is (1) see arguments, (2) see enviroment
    variables, (3) take boto3's default. Available arguments are:

    - ``aws_access_key_id``, ``AWS_ACCESS_KEY_ID``
    - ``aws_secret_access_key``, ``AWS_SECRET_ACCESS_KEY``
    - ``endpoint``, ``S3_ENDPOINT``

    '''

    def __init__(self, bucket, prefix=None,
                 endpoint=None, create_bucket=False,
                 aws_access_key_id=None,
                 aws_secret_access_key=None):
        self.bucket = bucket
        if prefix is not None:
            self.cwd = prefix
        else:
            self.cwd = ''

        # boto3.set_stream_logger()

        # import botocore
        # botocore.session.Session().set_debug_logger()

        kwargs = {}

        # IF these arguments are not defined, the library
        # automatically retrieves from AWS_ACCESS_KEY_ID and
        # AWS_SECRET_ACCESS_KEY.
        self.aws_access_key_id = aws_access_key_id
        if aws_access_key_id is not None:
            kwargs['aws_access_key_id'] = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        if aws_secret_access_key is not None:
            kwargs['aws_secret_access_key'] = aws_secret_access_key

        # We won't expect any enviroment variable for S3 endpoints
        # supported by boto3. Instead, we take S3_ENDPOINT in case
        # argument ``endpoint`` is not given. Otherwise, it goes to
        # boto3's default by giving ``None``.
        #
        # See also:
        # https://github.com/boto/boto3/issues/1375
        # https://github.com/boto/boto3/pull/2746
        self.endpoint = endpoint
        if self.endpoint is None:
            self.endpoint = os.getenv('S3_ENDPOINT')
        if self.endpoint is not None:
            kwargs['endpoint_url'] = self.endpoint

        # print('boto3.client options:', kwargs)
        self.client = boto3.client('s3', **kwargs)

        try:
            self.client.head_bucket(Bucket=bucket)
        except ClientError as e:
            if e.response['Error']['Code'] == '404' and create_bucket:
                res = self.client.create_bucket(Bucket=bucket)
                print("Bucket", bucket, "created:", res)
            else:
                raise e

    def open(self, path, mode='r', **kwargs):
        '''Opens an object accessor for read or write

        .. note:: Multi-part upload is not yet available.

        '''
        if 'a' in mode:
            raise io.UnsupportedOperation('Append is not supported')
        if 'r' in mode and 'w' in mode:
            raise io.UnsupportedOperation('Read-write mode is not supported')

        path = os.path.join(self.cwd, path)
        if 'r' in mode:
            obj = _ObjectReader(self.client, self.bucket, path, mode, kwargs)
            if 'b' in mode:
                obj = io.BufferedReader(obj)

        elif 'w' in mode:
            obj = _ObjectWriter(self.client, self.bucket, path, mode, kwargs)
            if 'b' in mode:
                obj = io.BufferedWriter(obj)

        else:
            raise RuntimeError(f'Unknown option: {mode}')

        return obj

    def list(self, prefix: str = "", recursive=False):
        '''List all objects (and prefixes)

        Although there is not concept of directory in AWS S3 API,
        common prefixes shows up like directories.

        '''
        self._checkfork()
        key = os.path.normpath(os.path.join(self.cwd, prefix))
        if key == '.':
            key = ''
        if key:
            key += '/'
        if '/../' in key or key.startswith('..'):
            raise ValueError('Invalid S3 key: {} as {}'.format(prefix, key))

        page_size = 1000
        paginator = self.client.get_paginator('list_objects_v2')
        paging_args = {
            'Bucket': self.bucket, 'Prefix': key,
            'PaginationConfig': {'PageSize': page_size}
        }
        if not recursive:
            paging_args['Delimiter'] = '/'

        iterator = paginator.paginate(**paging_args)
        for res in iterator:
            # print(res)
            for common_prefix in res.get('CommonPrefixes', []):
                yield common_prefix['Prefix']
            for content in res.get('Contents', []):
                yield content['Key'][len(key):]

    def stat(self, path):
        '''Imitate FileStat with S3 Object metadata

        '''
        self._checkfork()
        key = os.path.join(self.cwd, path)
        res = self.client.head_object(Bucket=self.bucket,
                                      Key=key)
        if res.get('DeleteMarker'):
            raise FileNotFoundError()

        return S3ObjectStat(key, res)

    def isdir(self, file_path: str):
        '''Does nothing

        .. note:: AWS S3 does not have concept of directory tree; what
           this function (and ``mkdir()`` and ``makedirs()`` should do
           and return? To be strict, it would be straightforward to
           raise ``io.UnsupportedOperation`` exception. But it just
           breaks users' applications that except quasi-compatible
           behaviour. Thus, imitating other file systems, like
           returning boolean or ``None`` would be nicer.

        '''
        # raise io.UnsupportedOperation("S3 doesn't have directory")
        pass

    def mkdir(self, file_path: str, mode=0o777, *args, dir_fd=None):
        '''Does nothing

        .. note:: see discussion in ``isdir()``.
        '''
        # raise io.UnsupportedOperation("S3 doesn't have directory")
        pass

    def makedirs(self, file_path: str, mode=0o777, exist_ok=False):
        '''Does nothing

        .. note:: see discussion in ``isdir()``.
        '''
        # raise io.UnsupportedOperation("S3 doesn't have directory")
        pass

    def exists(self, file_path: str):
        '''Returns the existence of objects

        For common prefixes, it does nothing. See discussion in ``isdir()``.
        '''
        self._checkfork()
        try:
            key = os.path.join(self.cwd, file_path)
            res = self.client.head_object(Bucket=self.bucket,
                                          Key=key)
            return not res.get('DeleteMarker')
        except ClientError as e:
            if e.response['Error']['Code'] == '404':
                return False
            else:
                raise e

    def rename(self, src, dst):
        '''Copies & removes the object

        Source and destination must be in the same bucket for
        ``pfio``, although AWS S3 supports inter-bucket copying.

        '''
        self._checkfork()
        source = {'Bucket': self.bucket, 'Key': os.path.join(self.cwd, src)}
        dst = os.path.join(self.cwd, dst)
        res = self.client.copy_object(Bucket=self.bucket,
                                      CopySource=source,
                                      Key=dst)
        if not res.get('CopyObjectResult'):
            # copy failed
            return
        return self.remove(source.get('Key'))

    def remove(self, file_path: str, recursive=False):
        '''Removes an object

        '''
        if recursive:
            raise io.UnsupportedOperation("Recursive delete not supported")

        self._checkfork()
        key = os.path.join(self.cwd, file_path)
        return self.client.delete_object(Bucket=self.bucket,
                                         Key=key)



from io import IOBase
from typing import Any, Callable, Iterator, Type

import pfio.cache  # NOQA
from pfio._context import DefaultContext
from pfio._typing import Optional, Union
from pfio.io import IO
from pfio.version import __version__  # NOQA

_DEFAULT_CONTEXT = DefaultContext()


def open_as_container(path: str) -> IO:
    """Opens a container and returns the handler

       Call the corresponding :func:`IO.open_as_container` upon the
       default handler.

       The ``path`` can be a POSIX path or a URI.

    """
    global _DEFAULT_CONTEXT
    default_context = _DEFAULT_CONTEXT

    return default_context.open_as_container(path)


def list(path_or_prefix: Optional[str] = None,
         recursive: bool = False) -> Iterator[str]:
    """Lists all the files and directories under the given ``path_or_prefix``

       Call the corresponding :func:`IO.list` upon the default handler.

       The ``path`` can be a POSIX path or a URI.

    """

    global _DEFAULT_CONTEXT
    default_context = _DEFAULT_CONTEXT

    if None is path_or_prefix:
        path_or_prefix = ""

    (handler, actual_path) = default_context.get_handler(path_or_prefix)
    return handler.list(actual_path, recursive)


def info() -> str:
    """Shows the detail of the current default handler

       Call the corresponding :func:`IO.info` upon the default handler.

    Returns:
        A string that describes the details of the default handler.

    """
    global _DEFAULT_CONTEXT
    default_context = _DEFAULT_CONTEXT

    (handler, dummy_path) = default_context.get_handler()
    return handler.info()


def open(file_path: str, mode: str = 'rb',
         buffering: int = -1, encoding: Optional[str] = None,
         errors: Optional[str] = None, newline: Optional[str] = None,
         closefd: bool = True,
         opener: Optional[Callable[
             [str, int], Any]] = None) -> Type['IOBase']:
    """Opens a regular file with ``mode``

    Call the corresponding :func:`IO.open` upon the default handler.

    If a POSIX path is given, the method use the default handler,
    and identifies the file from root path.
    See :func:`set_root` for details about root path and default handler.
    If a URI is given, right filesystem handler is automatically chosen
    by the library, according to the scheme included in the URI.

    """
    global _DEFAULT_CONTEXT
    default_context = _DEFAULT_CONTEXT

    (handler, actual_path) = default_context.get_handler(file_path)
    return handler.open(
        actual_path, mode, buffering, encoding,
        errors, newline, closefd, opener)


def set_root(uri_or_handler: Union[str, Type['IO']]) -> None:
    """Sets the current context to ``uri_or_handler``

    The context here refers to the default handler and the root path.
    The default handler points to a filesystem or a container which
    PFIO uses when called without a URI.
    Handlers can be created by :func:`create_handler` with
    the name of scheme. See the case 3 in the following example.

    Example::

        # Case 1
        # set_root by the name of uri
        pfio.set_root("posix")
        # open a file on posix filesystem with path "some/file"
        pfio.open("some/file")
        # override with a uri
        pfio.open("hdfs:///some/file/on/hdfs")

        # Case 2
        # set_root by uri
        pfio.set_root("hdfs:///some/directory")
        # open a file "on/hdfs" on hdfs under the " some/directory"
        pfio.open("on/hdfs")

        # Case 3
        handler = pfio.create_handler('hdfs')
        # set_root by handler
        pfio.set_root(handler)
        # open a file "/some/file/on/hdfs" on hdfs
        pfio.open("/some/file/on/hdfs")

        # Case 4
        handler = pfio.open_as_container('some.zip')
        # set_root by handler
        pfio.set_root(handler)
        # open a file "img.jpg" in "some.zip"
        pfio.open("img.jpg")

    The root path refers to a directory that PFIO works on.
    It is similar to current working directory, ``CWD`` in terms of the shell
    environment.
    The root path will only be set
    when the ``uri_or_handler`` points to a directory.
    Otherwise, it will be set to default,
    which represents the default working directory as follows:

    +---------+---------------------------+
    |         | Default Working Directory |
    +=========+===========================+
    | POSIX   | current working directory |
    +---------+---------------------------+
    | HDFS    | /user/``USERNAME`` [#f1]_ |
    +---------+---------------------------+
    | zip     |      top directory        |
    +---------+---------------------------+

    .. [#f1] For the details about the ``USERNAME`` in HDFS, please refer to \
    `HDFS Document <https://hadoop.apache.org/docs/current/\
            hadoop-project-dist/hadoop-hdfs/HdfsPermissionsGuide.html>`_


    Args:
        uri_or_handler (str or a object of a derived class of
            :class:`pfio.IO`): The ``uri_or_handler``
            can accept the following three kinds of values:

            1. the scheme (string): set the default handler
                according to the scheme.
                See :func:`create_handler` for supported scheme.

            2. a uri of directory (string): set the context
                to use the corresponding handler and set
                the root path to the given directory.

            3. a handler, which is an instance of
                :class:`pfio.IO`. Set the default handler
                to the given handler, and root path to default.

    """
    global _DEFAULT_CONTEXT
    default_context = _DEFAULT_CONTEXT

    default_context.set_root(uri_or_handler)


def create_handler(scheme: str) -> IO:
    """Returns a handler according to the given ``scheme``

    The current supported handlers are:

    1. posix
    2. hdfs

    See `scheme <https://github.com/pfnet/pfio/blob/master/\
            docs/source/design.rst#uri-expression-of-file-paths>`_
    for more details.


    Args:
        scheme (str): the name of the scheme.

    Returns:
        An object that implements the APIs defined in :class:`pfio.IO`.

    """
    global _DEFAULT_CONTEXT
    default_context = _DEFAULT_CONTEXT

    if not default_context.is_supported_scheme(scheme):
        raise ValueError("scheme {} is not supported".format(scheme))

    (handler, actual_path, is_URI) = \
        default_context.get_handler_by_name(scheme)
    return handler


def isdir(path: str) -> bool:
    """Returns ``True`` if the path is an existing directory

    Call the corresponding :func:`IO.isdir` upon the default handler.

    The ``path`` can be a POSIX path or a URI.
    """
    global _DEFAULT_CONTEXT
    default_context = _DEFAULT_CONTEXT

    (handler, actual_path, is_URI) = \
        default_context.get_handler_by_name(path)
    return handler.isdir(actual_path)


def mkdir(path: str, mode: int = 0o777, *,
          dir_fd: Optional[int] = None) -> None:
    """Makes a directory with mode

    Call the corresponding :func:`IO.mkdir` upon the default handler.

    The ``path`` can be a POSIX path or a URI.

    """
    global _DEFAULT_CONTEXT
    default_context = _DEFAULT_CONTEXT

    (handler, actual_path, is_URI) = \
        default_context.get_handler_by_name(path)
    return handler.mkdir(actual_path, mode, dir_fd=dir_fd)


def makedirs(path: str, mode: int = 0o777,
             exist_ok: bool = False) -> None:
    """Makes directories recursively with mode

    Call the corresponding :func:`IO.makedirs` upon the default handler.

    The ``path`` can be a POSIX path or a URI.

    """
    global _DEFAULT_CONTEXT
    default_context = _DEFAULT_CONTEXT

    (handler, actual_path, is_URI) = \
        default_context.get_handler_by_name(path)
    return handler.makedirs(actual_path, mode, exist_ok)


def exists(path: str) -> bool:
    """Returns ``True`` when the given ``path`` exists

    Call the corresponding :func:`IO.exists` upon the default handler.

    The ``path`` can be a POSIX path or a URI. or URIs.

    """
    global _DEFAULT_CONTEXT
    default_context = _DEFAULT_CONTEXT

    (handler, actual_path, is_URI) = \
        default_context.get_handler_by_name(path)
    return handler.exists(actual_path)


def rename(src: str, dst: str) -> None:
    """Renames the file from ``src`` to ``dst``

    Call the corresponding :func:`IO.rename` upon the default handler.

    Note the ``src`` and ``dst`` SHOULD be in the same filesystem.
    The ``src`` and ``dst`` can be either POSIX paths or URIs.

    """
    global _DEFAULT_CONTEXT
    default_context = _DEFAULT_CONTEXT

    (handler_src, actual_src, _is_URI1) = \
        default_context.get_handler_by_name(src)
    (handler_dst, actual_dst, _is_URI2) = \
        default_context.get_handler_by_name(dst)
    # TODO: containers are not supported here
    if type(handler_src) != type(handler_dst):
        raise NotImplementedError(
            "Moving between different systems is not supported")
    handler_src.rename(actual_src, actual_dst)


def remove(path: str, recursive: bool = False) -> None:
    """Removes a file or directory

    Call the corresponding :func:`IO.remove` upon the default handler.

    Args:
        path (str): the target path to remove. The ``path`` can be a
        regular file or a directory.
        The ``path`` can be a POSIX file path or a URI.

        recursive (bool): When the given path is a directory,
            all the files and directories under it will be removed.
            When the path is a file, this option is ignored.

    """
    global _DEFAULT_CONTEXT
    default_context = _DEFAULT_CONTEXT

    (handler, actual_path, is_URI) = \
        default_context.get_handler_by_name(path)
    return handler.remove(actual_path, recursive)


def get_root_dir() -> str:
    """get the current root path

    Returns:
        The current root path

    """
    global _DEFAULT_CONTEXT
    default_context = _DEFAULT_CONTEXT

    return default_context.get_root_dir()



import getpass
import io
import logging
import os
import re
import subprocess

import pyarrow
from pyarrow import hdfs

from .fs import FS, FileStat

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())


def _parse_principal_name_from_klist(output):
    output_array = output.split('\n')
    if len(output_array) < 2:
        return None

    principle_str = output_array[1]
    klist_principal_pattern = re.compile(
        r'Default principal: (?P<username>.+)@(?P<service>.+)')
    ret = klist_principal_pattern.match(principle_str)
    if ret:
        pattern_dict = ret.groupdict()
        return pattern_dict['username']
    else:
        return None


def _parse_principal_name_from_keytab(output):
    output_array = output.split('\n')
    if len(output_array) < 4:
        return None

    principle_str = output_array[3]
    keytab_principle_pattern = re.compile(
        r'\s+\d+ (?P<username>.+)@(?P<service>.+)')
    ret = keytab_principle_pattern.match(principle_str)
    if ret:
        pattern_dict = ret.groupdict()
        return pattern_dict['username']
    else:
        return None


def _get_principal_name_from_keytab():
    output = _run_klist(use_keytab=True)
    if output is None:
        return None

    return _parse_principal_name_from_keytab(output.decode('utf-8'))


def _get_principal_name_from_klist():
    output = _run_klist()
    if output is None:
        return None
    return _parse_principal_name_from_klist(output.decode('utf-8'))


def _run_klist(use_keytab=False):
    try:
        command = ['klist']
        if use_keytab:
            command += ['-k']
        pipe = subprocess.Popen(command, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        out, err = pipe.communicate()
        if out == b'' and err != b'':
            return None
        else:
            return out
    except OSError:
        # klist is not found
        return None


class HdfsFileStat(FileStat):
    """Detailed information of a file in HDFS

    Attributes:
        filename (str): Derived from `~FileStat`.
        last_modifled (float): Derived from `~FileStat`.
            No sub-second precision.
        last_accessed (float): UNIX timestamp of last access time.
            No sub-second precision.
        mode (int): Derived from `~FileStat`.
        size (int): Derived from `~FileStat`.
        owner (str): Owner of the file. Unlike `~PosixFileStat.owner`,
            this is a user name string instead of an integer.
        group (str): Group of the file in string.
        replication (int): Number of replications of the file in HDFS.
        block_size (int): Block size in bytes.
        kind (str): Group of the file in string.
    """

    def __init__(self, info):
        mode = info['permissions']
        if info['kind'] == 'file':
            mode |= 0o100000
        elif info['kind'] == 'directory':
            mode |= 0o40000

        self.filename = info['path']
        self.mode = mode
        self.last_modified = float(info['last_modified'])
        self.last_accessed = float(info['last_accessed'])
        for k in ('size', 'owner', 'group', 'replication',
                  'block_size', 'kind'):
            setattr(self, k, info[k])


class Hdfs(FS):
    def __init__(self, cwd=None):
        super().__init__()
        self.connection = hdfs.connect()
        assert self.connection is not None
        self.username = self._get_principal_name()
        self.cwd = os.path.join('/user', self.username)

        if cwd:
            self.cwd = os.path.join(self.cwd, cwd)

        # set nameservice
        _file_in_root = self.connection.ls("/")[0]
        self.nameservice = _file_in_root[:_file_in_root.rfind("/")]

    def _get_principal_name(self):
        # get the default principal name from `klist` cache
        principal_name = _get_principal_name_from_klist()

        if principal_name is not None:
            return principal_name

        # try getting principal name from keytab
        principal_name = _get_principal_name_from_keytab()
        if principal_name is not None:
            return principal_name

        # in case every thing, use the login username instead
        return self._get_login_username()

    def _get_login_username(self):
        return getpass.getuser()

    def open(self, file_path, mode='rb',
             buffering=-1, encoding=None, errors=None,
             newline=None, closefd=True, opener=None):
        self._checkfork()
        path = os.path.join(self.cwd, file_path)
        orig_mode = mode

        # hdfs only support open in 'b'
        if 'b' not in mode:
            mode += 'b'
        try:
            file_obj = self.connection.open(path, mode)

        except pyarrow.lib.ArrowIOError as e:
            raise IOError("open file error :{}".format(str(e)))

        if 'b' not in orig_mode:
            file_obj = io.TextIOWrapper(file_obj, encoding, errors, newline)
        elif 'r' in orig_mode:
            # Wrap file_obj with io.BufferedReader for ``peek()``, to
            # significiantly improve unpickle performance.
            file_obj = io.BufferedReader(file_obj)
        else:
            file_obj = io.BufferedWriter(file_obj)

        return file_obj

    def subfs(self, rel_path):
        return Hdfs(os.path.join(self.cwd, rel_path))

    def close(self):
        self._checkfork()
        self.connection.close()

    def list(self, path_or_prefix: str = "", recursive=False):
        self._checkfork()
        path_or_prefix = os.path.join(self.cwd, path_or_prefix)

        target_dir = self.connection.info(path_or_prefix)
        if target_dir['kind'] != "directory":
            raise NotADirectoryError(path_or_prefix)

        target_dir_path = target_dir['path']
        # +1 to include the "/"
        full_uri_prefix_offset = len(target_dir_path) + 1

        if recursive:
            yield from self._recursive_list(full_uri_prefix_offset,
                                            path_or_prefix)
        else:
            dir_list = self.connection.ls(path_or_prefix)
            for _dir in dir_list:
                yield os.path.basename(_dir)

    def _recursive_list(self, full_uri_prefix_offset, path):
        for _file in self.connection.ls(path, detail=True):
            file_name = _file['name']
            # convert the full URI to relative path from path_or_prefix
            # to align with posix
            # e.g. "hdfs://nameservice/prefix_dir/testfile"
            # => "testfile"
            yield file_name[full_uri_prefix_offset:]

            if 'directory' == _file['kind']:
                yield from self._recursive_list(full_uri_prefix_offset,
                                                file_name)

    def stat(self, path):
        self._checkfork()
        path = os.path.join(self.cwd, path)
        return HdfsFileStat(self.connection.info(path))

    def isdir(self, path: str):
        path = os.path.join(self.cwd, path)
        return self.stat(path).isdir()

    def mkdir(self, path: str, *args, dir_fd=None):
        self._checkfork()
        path = os.path.join(self.cwd, path)
        return self.connection.mkdir(path)

    def makedirs(self, file_path: str, mode=0o777, exist_ok=False):
        file_path = os.path.join(self.cwd, file_path)
        return self.mkdir(file_path, mode, exist_ok)

    def exists(self, file_path: str):
        self._checkfork()
        file_path = os.path.join(self.cwd, file_path)
        return self.connection.exists(file_path)

    def rename(self, src, dst):
        self._checkfork()
        s = os.path.join(self.cwd, src)
        d = os.path.join(self.cwd, dst)
        return self.connection.rename(s, d)

    def remove(self, path, recursive=False):
        self._checkfork()
        delpath = os.path.join(self.cwd, path)
        return self.connection.delete(delpath, recursive)



import warnings

from pfio.chainer_extensions.snapshot import load_snapshot  # NOQA

warnings.warn("Chainer extentions are deprecated and "
              "will be removed. Please use 'pfio' instead.",
              DeprecationWarning)



import sys
import warnings

warnings.warn("package 'chainerio' is deprecated and will be removed."
              " Please use 'pfio' instead.",
              DeprecationWarning)

# make sure pfio is in sys.modules
import pfio  # NOQA

sys.modules[__name__] = __import__('pfio')
